import click
from .shared_params import (
    use_common_state,
    project_id_arg,
    sample_ids_arg,
    yes_option,
    private_option,
    handle_project_id,
    handle_multiple_sample_ids,
)
from geoseeq import Organization


@click.group('manage')
def cli_manage():
    """Manage GeoSeeq objects, add, delete, etc."""
    pass


@cli_manage.command('add-samples-to-project')
@use_common_state
@yes_option
@project_id_arg
@sample_ids_arg
def cli_add_samples(state, yes, project_id, sample_ids):
    """Add samples to a project.

    ---

    Example Usage:

    \b
    # Add two samples to "My Org/My Project"
    $ geoseeq manage add-samples-to-project "My Org/My Project" c8141b43-398a-4ab5-a8cf-0aa1fd902a95 553ba5fc-5f8c-45a4-a057-3cf4b310c53d

    \b
    # Add all samples in "My Org/My Source Project" to "My Org/My Target Project"
    $ geoseeq manage add-samples-to-project "My Org/My Target Project" "My Org/My Source Project"

    ---

    Command Arguments:

    \b
    [PROJECT_ID] is the name or ID of the project to add samples to.

    \b
    [SAMPLE_IDS]... can be a list of sample names or IDs, files containing a list of sample names or IDs, or a mix of both.
    The first element in the list can optionally be a project ID or name.
    If a project ID is not provided, then sample ids must be UUIDs or GRNs, not names.
    If only a project ID is provided, then all samples in the project will be added.

    ---
    """
    knex = state.get_knex()
    proj = handle_project_id(knex, project_id, yes=yes)
    samples = handle_multiple_sample_ids(knex, sample_ids)
    click.echo(f'Adding {len(samples)} samples to {proj.name}', err=True)
    if not yes:
        click.confirm('Continue?', abort=True)
    for sample in samples:
        proj.add_sample(sample)
    proj.save()
    click.echo(f'Added {len(samples)} to {proj.name}', err=True)


@cli_manage.command('create-org')
@use_common_state
@click.option('--exists-ok/--no-exists-ok', default=True)
@yes_option
@click.argument('org_name')
def cli_create_org(state, exists_ok, yes, org_name):
    """Create an organization.
    
    ---

    Example Usage:

    \b
    # Create an organization named "My Org"
    $ geoseeq manage create-org "My Org"

    ---

    Command Arguments:

    \b
    [ORG_NAME] is the name of the organization to create.

    ---
    """
    knex = state.get_knex()
    if yes or click.confirm(f'Create Organization {org_name}?'):
        if exists_ok:
            org = Organization(knex, org_name).idem()
        else:
            org = Organization(knex, org_name).create()
        click.echo(f'Created Organization {org}', err=True)


@cli_manage.command('create-project')
@use_common_state
@yes_option
@private_option
@project_id_arg
def cli_create_project(state, yes, private, project_id):
    """Create a project.

    ---
    
    Example Usage:
    
    \b
    # Create a project named "My Project" in "My Org"
    $ geoseeq manage create-project "My Org/My Project"

    ---

    Command Arguments:

    \b
    [PROJECT_ID] is the name or ID of the project to create.

    ---
    """
    knex = state.get_knex()
    proj = handle_project_id(knex, project_id, yes=yes, private=private)
    click.echo(f'Created Project {proj}', err=True)


@cli_manage.command('create-samples')
@use_common_state
@yes_option
@private_option
@project_id_arg
@click.argument('sample_names', nargs=-1)
def cli_create_samples(state, yes, private, project_id, sample_names):
    """Create samples in the specified project.
    
    ---
    
    Example Usage:
    
    \b
    # Create two samples in "My Org/My Project"
    $ geoseeq manage create-samples "My Org/My Project" "Sample 1" "Sample 2"

    ---

    Command Arguments:

    \b
    [PROJECT_ID] is the name or ID of the project to create samples in.

    \b
    [SAMPLE_NAMES]... is a list of sample names to create.

    ---
    """
    knex = state.get_knex()
    proj = handle_project_id(knex, project_id, yes=yes, private=private)
    for sample_name in sample_names:
        sample = proj.sample(sample_name).idem()
        click.echo(f'Created Sample {sample}', err=True)


@cli_manage.command('delete-samples')
@use_common_state
@yes_option
@sample_ids_arg
def cli_delete_samples(state, yes, sample_ids):
    """Delete samples by their ID or name.
    
    Confirm before deleting unless --yes is provided.

    ---

    Example Usage:

    \b
    # Delete two samples by ID
    $ geoseeq manage delete-samples c8141b43-398a-4ab5-a8cf-0aa1fd902a95 553ba5fc-5f8c-45a4-a057-3cf4b310c53d

    \b
    # Delete one samples by name
    $ geoseeq manage delete-samples "My Org/My Project" "My Sample"

    ---

    Command Arguments:

    \b
    [SAMPLE_IDS]... can be a list of sample names or IDs, files containing a list of sample names or IDs, or a mix of both.
    The first element in the list can optionally be a project ID or name.
    If a project ID is not provided, then sample ids must be UUIDs or GRNs, not names.
    If only a project ID is provided, then all samples in the project will be deleted.

    ---
    """
    knex = state.get_knex()
    samples = handle_multiple_sample_ids(knex, sample_ids)
    click.echo(f'Deleting {len(samples)} samples', err=True)
    if not yes:
        click.confirm('Continue?', abort=True)
    for sample in samples:
        sample.delete()
    click.echo(f'Deleted {len(samples)} samples', err=True)



from geoseeq import GeoseeqNotFoundError
from .from_uuids import (
    org_from_uuid,
    project_from_uuid,
    sample_from_uuid,
    sample_result_folder_from_uuid,
    project_result_folder_from_uuid,
    result_folder_from_uuid,
    sample_result_file_from_uuid,
    project_result_file_from_uuid,
    result_file_from_uuid,
)
from .from_names import (
    org_from_name,
    project_from_name,
    sample_from_name,
    sample_result_folder_from_name,
    project_result_folder_from_name,
    result_folder_from_name,
    sample_result_file_from_name,
    project_result_file_from_name,
    result_file_from_name,
)
from .utils import is_name, is_grn


def resolve_id(knex, id):
    """Return the object which the id points to."""
    if is_grn(id):
        return resolve_grn(knex, id)
    if is_name(id):
        return resolve_name(knex, id)
    raise ValueError(f'"{id}" is not a GRN, or name. UUIDs cannot be resolved without a type')


def resolve_name(knex, name):
    """Return the object which the name points to."""
    assert is_name(name), f'"{name}" is not a name'
    tkns = name.split("/")
    if len(tkns) == 1:  # org
        return "org", org_from_name(knex, name)
    if len(tkns) == 2:  # project
        return "project", project_from_name(knex, name)
    if len(tkns) == 3:  # sample or project result folder
        try:
            return "sample", sample_from_name(knex, name)
        except GeoseeqNotFoundError:
            return "folder", result_folder_from_name(knex, name)
    if len(tkns) == 4:  # sample result folder or project result file
        try:
            return "folder", result_folder_from_name(knex, name)
        except GeoseeqNotFoundError:
            return "file", result_file_from_name(knex, name)
    if len(tkns) == 5:  # sample result file
        return "file", result_file_from_name(knex, name)
    raise GeoseeqNotFoundError(f'Name "{name}" not found')


def resolve_grn(knex, grn):
    """Return the object which the grn points to."""
    assert is_grn(grn), f'"{grn}" is not a GRN'
    _, _, object_type, uuid = grn.split(':')
    if object_type == 'org':
        return object_type, org_from_uuid(knex, uuid)
    if object_type == 'project':
        return object_type, project_from_uuid(knex, uuid)
    if object_type == 'sample':
        return object_type, sample_from_uuid(knex, uuid)
    if object_type == 'folder':
        return object_type, result_folder_from_uuid(knex, uuid)
    if object_type == 'file':
        return object_type, result_file_from_uuid(knex, uuid)
    raise GeoseeqNotFoundError(f'Type "{object_type}" not found')


import click
from .shared_params import (
    use_common_state,
)
from geoseeq.search import Search



@click.command('search')
@use_common_state
@click.option('--sep', default='\t')
@click.argument('search_terms', nargs=-1)
def cli_search(state, sep, search_terms):
    """Run a search and write out a list of sample IDs.
    
    ---

    Example Usage:

    \b
    # Search for samples in London
    $ geoseeq search 'city: london'

    \b
    # Search for samples in london with Staphylolococcus
    $ geoseeq search 'city: london' 'staphylococcus'

    ---

    Command Arguments:

    \b
    [SEARCH_TERMS]... a list of search terms to search for

    ---
    """
    search = Search(state.get_knex())
    for search_term in search_terms:
        search.add_search_term(search_term)
    search.run_search()
    print(search.sample_table().to_csv(sep=sep), file=state.outfile)


from geoseeq.utils import md5_checksum

class Checksum:

    def __init__(self, value, method):
        self.value = value
        self.method = method.lower()
    
    def verify(self, path):
        """Return True iff the checksum for path matches the stored checksum."""
        if self.method == 'none':
            return False
        if self.method == 'md5':
            return md5_checksum(path) == self.value
        raise NotImplementedError(f'Mehtod "{self.method}" not supported')

    def to_blob(self):
        return {
            "value": self.value,
            "method": self.method,
        }

    @classmethod
    def from_blob(cls, blob):
        return cls(blob['value'], blob['method'])




import uuid
from geoseeq.blob_constructors import (
    project_from_uuid,
    org_from_uuid,
    sample_from_uuid,
    sample_result_folder_from_uuid,
    project_result_folder_from_uuid,
)
from geoseeq.knex import GeoseeqNotFoundError
from os.path import isfile
from geoseeq.id_constructors.utils import is_grn, is_uuid, is_grn_or_uuid
from .obj_getters import (
    _get_org,
    _get_org_and_proj,
    _get_org_proj_and_sample,
    _get_org_proj_sample_and_folder,
    _get_org_proj_and_folder,
)

    

def read_els_from_file(file):
    """Return a list of els from a file"""
    with open(file) as f:
        return [line for line in (l.strip() for l in f) if line]
    

def flatten_list_of_els_and_files(els):
    """Return a list of els with any files replaced by the contents of the file"""
    flattened = []
    for el in els:
        if isfile(el):
            flattened.extend(read_els_from_file(el))
        else:
            flattened.append(el)
    return flattened


def el_is_org_id(knex, el):
    """Return an organization object if `el` is an organization ID, otherwise return None"""
    if is_uuid(el):
        try:
            org = org_from_uuid(knex, el)
            return org
        except GeoseeqNotFoundError:
            return None
    elif is_grn(el):
        if ':organization:' in el:
            return org_from_uuid(knex, el.split(':')[-1])
        return None
    return None


def el_is_project_id(knex, el):
    """Return a project object if `el` is a project ID, otherwise return None"""
    try:
       return handle_project_id(knex, el)
    except (ValueError, GeoseeqNotFoundError):
        return None
    

def handle_org_id(knex, org_id, yes=False, create=True):
    """Return an organization object.
    
    Organization ID must be one of the following types:
    - a UUID
    - an organization name
    - a GeoSeeq Resource Number (GRN)

    If the organization name is provided, the organization will be created if it does not exist
    """
    if is_grn_or_uuid(org_id):
        org_id = org_id.split(':')[-1]  # this gives a UUID either way
        org = org_from_uuid(knex, org_id)
        return org
    org = _get_org(knex, org_id, yes=yes, create=create)
    return org
    

def handle_project_id(knex, project_id, yes=False, private=True, create=True):
    """Return a project object
    
    Project ID must be one of the following types:
    - a UUID
    - an organization name and project name separated by a slash
    - a GeoSeeq Resource Number (GRN)

    If both an organization name and project name are provided, the project and organization will
    be created if they do not exist and the user confirms.
    """

    if '/' in project_id:
        org_name, project_name = project_id.split('/')
        _, project = _get_org_and_proj(knex, org_name, project_name, yes=yes, private=private, create=create)
        return project
    elif is_grn_or_uuid(project_id):
        proj_uuid = project_id.split(':')[-1]  # this gives a UUID either way
        project = project_from_uuid(knex, proj_uuid)
        return project
    raise ValueError('project_id must be a UUID, an organization name and project name, or a GRN')


def handle_folder_id(knex, folder_id, yes=False, private=True, create=True):
    """Return a folder object, creating it if it does not exist and the user confirms

    Can be either a project or sample folder depending on context
    """
    if '/' in folder_id:
        tkns = folder_id.split('/')
        if len(tkns) == 4:  # org, proj, sample, folder
            org_name, project_name, sample_name, folder_name = tkns
            _, _, _, folder = _get_org_proj_sample_and_folder(knex, org_name, project_name, sample_name, folder_name, yes=yes, private=private, create=create)
        elif len(tkns) == 3:  # proj, sample, folder
            project_name, sample_name, folder_name = tkns
            _, _, folder = _get_org_proj_and_folder(knex, project_name, sample_name, folder_name, yes=yes, private=private, create=create)
        else:
            raise ValueError(f'Invalid folder ID: {folder_id}')
        return folder
    elif is_grn_or_uuid(folder_id):
        folder_uuid = folder_id.split(':')[-1]  # this gives a UUID either way
        # we guess that this is a sample folder, TODO: use GRN if available
        try:
            folder = sample_result_folder_from_uuid(knex, folder_uuid)
        except GeoseeqNotFoundError:
            folder = project_result_folder_from_uuid(knex, folder_uuid)
        return folder
    raise ValueError('sample_folder_id must be a UUID, an organization name and project name, or a GRN')


def handle_multiple_sample_ids(knex, sample_ids, proj=None):
    """Return a list of fetched sample objects
    
    `sample_ids` may have three different structures:
     - the sole element may be a project ID of any type, samples will be fetched from that project
     - the first element may be a project ID of any type, followed by a list of sample ids of any type or project
     - No project ID is provided, and each element is a sample id of any type

    Any sample may in fact be a file containing sample IDs, in which case the file will be read line by line
    and each element will be a sample ID

    If `one_project` is True, all samples must be from the same project
    """
    project_as_arg = bool(proj)
    if proj or (proj := el_is_project_id(knex, sample_ids[0])):
        # The first element is a project ID. Remaining els may be sample IDs or names
        if not project_as_arg and sample_ids:
            sample_ids = list(sample_ids)[1:]
        if len(sample_ids) == 0:
            return list(proj.get_samples(cache=False))
        else:
            samples = []
            for el in flatten_list_of_els_and_files(sample_ids):
                if is_grn_or_uuid(el):
                    el = el.split(':')[-1]
                    samples.append(sample_from_uuid(knex, el))
                else:  # assume it's a sample name
                    samples.append(proj.sample(el).get())
    else:
        # No project ID provided. Each element is a sample ID
        samples = []
        for el in flatten_list_of_els_and_files(sample_ids):
            if is_grn_or_uuid(el):
                el = el.split(':')[-1]
                samples.append(sample_from_uuid(knex, el))
            else:
                raise ValueError(f'"{el}" is not a valid sample ID. To use samples by name, provide a project ID first.')
    return samples





"""Test suite for plotting library."""
import random
from os import environ
from os.path import dirname, join
from unittest import TestCase, skip

from geoseeq.plotting.map import Map


class TestGeoseeqPlotting(TestCase):
    """Test suite for packet building."""

    def test_make_map_complex(self):
        """Test that we can create a map and turn it into a dict."""
        map = Map()\
            .set_center(0, 0)\
            .set_zoom(2)\
            .add_light_base_map()\
            .add_administrative_overlay()\
            .add_places_overlay()
        map.to_dict()

    def test_make_map_simple(self):
        """Test that we can create a map and turn it into a dict."""
        map = Map()\
            .add_light_base_map()
        map.to_dict()
        
        


from .result import ProjectResultFolder
from .remote_object import RemoteObject
from .sample import Sample
from .utils import paginated_iterator
import json
import pandas as pd



class Project(RemoteObject):
    remote_fields = [
        "uuid",
        "created_at",
        "name",
        "privacy_level",
        "description",
    ]
    optional_remote_fields = [
        "privacy_level",
    ]
    parent_field = "org"
    url_prefix = "sample_groups"

    def __init__(
        self,
        knex,
        org,
        name,
        metadata={},
        is_library=True,
        is_public=False,
        storage_provider="default",
        privacy_level=None,
    ):
        super().__init__(self)
        self.knex = knex
        self.org = org
        self.new_org = None
        self.name = name
        self.is_library = is_library
        self.is_public = is_public
        self._sample_cache = []
        self._deleted_sample_cache = []
        self._get_sample_cache = []
        self._get_result_cache = []
        self.metadata = metadata
        self.storage_provider = storage_provider
        self.privacy_level = privacy_level

    def change_org(self, org):
        self.new_org = org
        self._modified = True

    def get_post_data(self):
        data = {field: getattr(self, field) for field in self.remote_fields if hasattr(self, field)}
        data["organization"] = self.org.uuid
        data['description'] = self.description if hasattr(self, 'description') and self.description else self.name
        data['privacy_level'] = self.privacy_level if hasattr(self, 'privacy_level') and self.privacy_level else 'private'
        data['storage_provider_name'] = self.storage_provider
        if self.new_org:
            if isinstance(self.new_org, RemoteObject):
                data["organization"] = self.new_org.uuid
            else:
                data["organization"] = self.new_org
        if data["uuid"] is None:
            data.pop("uuid")
        return data

    def nested_url(self):
        return self.org.nested_url() + f"/sample_groups/{self.name}"

    def _save_group_obj(self):
        data = self.get_post_data()
        url = f"sample_groups/{self.uuid}"
        self.knex.put(url, json=data)

    def _save_sample_list(self):
        sample_uuids = []
        for sample_uuid in self._sample_cache:
            sample_uuids.append(sample_uuid)
        if sample_uuids:
            url = f"sample_groups/{self.uuid}/samples"
            chunk_size = 100
            for i in range(0, len(sample_uuids), chunk_size):
                self.knex.post(url, json={"sample_uuids": sample_uuids[i : i + chunk_size]})
        self._sample_cache = []

    def _delete_sample_list(self):
        sample_uuids = []
        for sample_uuid in self._deleted_sample_cache:
            sample_uuids.append(sample_uuid)
        if sample_uuids:
            url = f"sample_groups/{self.uuid}/samples"
            self.knex.delete(url, json={"sample_uuids": sample_uuids})
        self._deleted_sample_cache = []

    def _save(self):
        self._save_group_obj()
        self._save_sample_list()
        self._delete_sample_list()

    def _get(self, allow_overwrite=False):
        """Fetch the result from the server."""
        self.org.idem()
        blob = self.get_cached_blob()
        if not blob:
            blob = self.knex.get(self.nested_url())
            self.load_blob(blob, allow_overwrite=allow_overwrite)
            self.cache_blob(blob)
        else:
            self.load_blob(blob)

    def _create(self):
        self.org.idem()
        post_data = self.get_post_data()
        blob = self.knex.post(
            f"sample_groups?format=json",
            json=post_data,
        )
        self.load_blob(blob)
    
    def add_sample_uuids(self, sample_uuids):
        """Return this group and add a sample to this group.

        Do not contact server until `.save()` is called on this group.
        """
        for sample_uuid in sample_uuids:
            self._sample_cache.append(sample_uuid)
        self._modified = True
        return self

    def add_sample(self, sample):
        """Return this group and add a sample to this group.

        Do not contact server until `.save()` is called on this group.
        """
        self._sample_cache.append(sample.uuid)
        self._modified = True
        return self

    def remove_sample(self, sample):
        """Return this group and remove a sample to this group.

        Do not contact server until `.save()` is called on this group.
        """
        self._deleted_sample_cache.append(sample.uuid)
        self._modified = True
        return self

    def sample(self, sample_name, metadata={}):
        return Sample(self.knex, self, sample_name, metadata=metadata)

    def result_folder(self, module_name, replicate=None, metadata={}):
        """Return a ProjectResultFolder object for this project."""
        return ProjectResultFolder(
            self.knex, self, module_name, replicate=replicate, metadata=metadata
        )

    def analysis_result(self, *args, **kwargs):
        """Return a ProjectResultFolder object for this project.
        
        Alias for result_folder."""
        return self.result_folder(*args, **kwargs)

    def get_samples(self, cache=True, error_handler=None):
        """Yield samples fetched from the server."""
        if cache and self._get_sample_cache:
            for sample in self._get_sample_cache:
                yield sample
            return
        url = f"sample_groups/{self.uuid}/samples"
        for sample_blob in paginated_iterator(self.knex, url, error_handler=error_handler):
            sample = self.sample(sample_blob["name"])
            sample.load_blob(sample_blob)
            sample.cache_blob(sample_blob)
            # We just fetched from the server so we change the RemoteObject
            # meta properties to reflect that
            sample._already_fetched = True
            sample._modified = False
            if cache:
                self._get_sample_cache.append(sample)
            else:
                yield sample
        if cache:
            for sample in self._get_sample_cache:
                yield sample

    def get_sample_uuids(self, cache=True, error_handler=None):
        """Yield samples uuids fetched from the server."""
        if cache and self._get_sample_cache:
            for sample in self._get_sample_cache:
                yield sample.uuid
            return
        url = f"sample_groups/{self.uuid}/samples"
        for sample_blob in paginated_iterator(self.knex, url, error_handler=error_handler):
            yield sample_blob['uuid']

    def get_analysis_results(self, cache=True):
        """Yield ProjectResultFolder objects for this project fetched from the server.
        
        Alias for get_result_folders."""
        return self.get_result_folders(cache=cache)

    def get_result_folders(self, cache=True):
        """Yield ProjectResultFolder objects for this project fetched from the server."""
        if cache and self._get_result_cache:
            for ar in self._get_result_cache:
                yield ar
            return
        url = f"sample_group_ars?sample_group_id={self.uuid}"
        result = self.knex.get(url)
        for result_blob in result["results"]:
            result = self.analysis_result(result_blob["module_name"])
            result.load_blob(result_blob)
            # We just fetched from the server so we change the RemoteObject
            # meta properties to reflect that
            result._already_fetched = True
            result._modified = False
            if cache:
                self._get_result_cache.append(result)
            else:
                yield result
        if cache:
            for ar in self._get_result_cache:
                yield ar

    def get_manifest(self):
        """Return a manifest for this group."""
        url = f"sample_groups/{self.uuid}/manifest"
        return self.knex.get(url)

    def get_module_counts(self):
        """Return a dictionary with module counts for samples in this group."""
        url = f"sample_groups/{self.uuid}/module_counts"
        return self.knex.get(url)

    def get_sample_metadata(self):
        """Return a pandas dataframe with sample metadata."""
        url = f"sample_groups/{self.uuid}/metadata"
        blob = self.knex.get(url)
        return pd.DataFrame.from_dict(blob, orient="index")

    def __str__(self):
        return f"<Geoseeq::Project {self.name} {self.uuid} />"

    def __repr__(self):
        return f"<Geoseeq::Project {self.name} {self.uuid} />"

    def pre_hash(self):
        return "PROJ" + self.name + self.org.pre_hash()

SampleGroup = Project  # alias for backwards compatibility


from geoseeq import GeoseeqNotFoundError

from .from_blobs import *


def org_from_uuid(knex, uuid):
    """Return the organization object which the uuid points to."""
    blob = knex.get(f"organizations/{uuid}")
    org = org_from_blob(knex, blob)
    return org


def project_from_uuid(knex, uuid):
    """Return the project object which the uuid points to."""
    blob = knex.get(f"sample_groups/{uuid}")
    project = project_from_blob(knex, blob)
    return project


sample_group_from_uuid = project_from_uuid  # Alias


def sample_from_uuid(knex, uuid):
    """Return the sample object which the uuid points to."""
    blob = knex.get(f"samples/{uuid}")
    sample = sample_from_blob(knex, blob)
    return sample


def sample_result_folder_from_uuid(knex, uuid):
    """Return the sample result folder object which the uuid points to."""
    blob = knex.get(f"sample_ars/{uuid}")
    ar = sample_result_folder_from_blob(knex, blob)
    return ar


sample_ar_from_uuid = sample_result_folder_from_uuid  # Alias


def project_result_folder_from_uuid(knex, uuid):
    """Return the project result folder object which the uuid points to."""
    blob = knex.get(f"sample_group_ars/{uuid}")
    ar = project_result_folder_from_blob(knex, blob)
    return ar


sample_group_ar_from_uuid = project_result_folder_from_uuid  # Alias


def result_folder_from_uuid(knex, uuid):
    """Return a result folder object which the uuid points to.

    Guess the result folder is a sample result folder. If not, try a project result folder.
    """
    try:
        return sample_result_folder_from_uuid(knex, uuid)
    except GeoseeqNotFoundError:
        return project_result_folder_from_uuid(knex, uuid)


def sample_result_file_from_uuid(knex, uuid):
    """Return the sample result file object which the uuid points to."""
    blob = knex.get(f"sample_ar_fields/{uuid}")
    ar = sample_result_file_from_blob(knex, blob)
    return ar


sample_ar_field_from_uuid = sample_result_file_from_uuid  # Alias


def project_result_file_from_uuid(knex, uuid):
    """Return the project result file object which the uuid points to."""
    blob = knex.get(f"sample_group_ar_fields/{uuid}")
    ar = project_result_file_from_blob(knex, blob)
    return ar


sample_group_ar_field_from_uuid = project_result_file_from_uuid  # Alias


def result_file_from_uuid(knex, uuid):
    """Return a result file object which the uuid points to.

    Guess the result file is a sample result file. If not, try a project result file.
    """
    try:
        return sample_result_file_from_uuid(knex, uuid)
    except GeoseeqNotFoundError:
        return project_result_file_from_uuid(knex, uuid)


def pipeline_run_from_uuid(knex, uuid):
    """Return a pipeline run object which the uuid points to."""
    blob = knex.get(f"app_runs/{uuid}")
    pipeline_run = pipeline_run_from_blob(knex, blob)
    return pipeline_run



"""Construct objects from names.

This module is used to construct objects from names. It is used by the

Names are human-readable ids that are used to identify objects in GeoSeeq.
They similar to file paths, e.g.:
 - "My Org"
 - "My Org/My Project"
 - "My Org/My Project/My Sample/My Result Folder/My Result File"
 - "My Org/My Project/My Result Folder/My Result File"

Unlike UUIDs or GRNs names can be changed by the user.
This makes them more human-readable, but not reliable for permanent references.
"""
from geoseeq.organization import Organization
from geoseeq import GeoseeqNotFoundError


def org_from_name(knex, name):
    """Return the organization object which the name points to.
    
    e.g. "My Org"
    """
    tkns = name.split("/")
    org_name = tkns[0]
    org = Organization(knex, org_name)
    org.get()
    return org


def project_from_name(knex, name):
    """Return the project object which the name points to.
    
    e.g. "My Org/My Project"
    """
    tkns = name.split("/")
    proj_name = tkns[1]
    org = org_from_name(knex, name)
    proj = org.project(proj_name)
    proj.get()
    return proj


def sample_from_name(knex, name):
    """Return the sample object which the name points to.
    
    e.g. "My Org/My Project/My Sample"
    """
    tkns = name.split("/")
    sample_name = tkns[2]
    proj = project_from_name(knex, name)
    sample = proj.sample(sample_name)
    sample.get()
    return sample


def sample_result_folder_from_name(knex, name):
    """Return the sample result folder object which the name points to.
    
    e.g. "My Org/My Project/My Sample/My Result Folder"
    """
    tkns = name.split("/")
    result_folder_name = tkns[3]
    sample = sample_from_name(knex, name)
    ar = sample.result_folder(result_folder_name)
    ar.get()
    return ar


def project_result_folder_from_name(knex, name):
    """Return the project result folder object which the name points to.
    
    e.g. "My Org/My Project/My Result Folder"
    """
    tkns = name.split("/")
    result_folder_name = tkns[2]
    proj = project_from_name(knex, name)
    rf = proj.result_folder(result_folder_name)
    rf.get()
    return rf


def result_folder_from_name(knex, name):
    """Return a result folder object which the name points to.
    
    Guess the result folder is a sample result folder. If not, try a project result folder.
    """
    tkns = name.split("/")
    if len(tkns) >= 4:
        try:
            return sample_result_folder_from_name(knex, name)
        except GeoseeqNotFoundError:
            return project_result_folder_from_name(knex, name)
    else:  # can't be a sample result folder
        return project_result_folder_from_name(knex, name)


def sample_result_file_from_name(knex, name):
    """Return the sample result file object which the name points to.
    
    e.g. "My Org/My Project/My Sample/My Result Folder/My Result File"
    """
    tkns = name.split("/")
    result_file_name = tkns[4]
    r_folder = sample_result_folder_from_name(knex, name)
    r_file = r_folder.result_file(result_file_name)
    r_file.get()
    return r_file


def project_result_file_from_name(knex, name):
    """Return the project result file object which the name points to.
    
    e.g. "My Org/My Project/My Result Folder/My Result File"
    """
    tkns = name.split("/")
    result_file_name = tkns[3]
    r_folder = project_result_folder_from_name(knex, name)
    r_file = r_folder.result_file(result_file_name)
    r_file.get()
    return r_file


def result_file_from_name(knex, name):
    """Return a result file object which the name points to.
    
    Guess the result file is a sample result file. If not, try a project result file.
    """
    tkns = name.split("/")
    if len(tkns) >= 5:
        try:
            return sample_result_file_from_name(knex, name)
        except GeoseeqNotFoundError:
            return project_result_file_from_name(knex, name)
    else:  # can't be a sample result file
        return project_result_file_from_name(knex, name)




"""Test suite for experimental functions."""
import random
import sys
from os import environ
from unittest import TestCase, skip

from geoseeq import (
    Knex,
    Organization,
    GeoseeqOtherError,
    GeoseeqNotFoundError,
)
from requests.exceptions import ConnectionError

ENDPOINT = environ.get("GEOSEEQ_API_TESTING_ENDPOINT", "http://127.0.0.1:8000")
TOKEN = environ.get("GEOSEEQ_API_TOKEN", "<no_token>")


def random_str(len=12):
    """Return a random alphanumeric string of length `len`."""
    out = random.choices("abcdefghijklmnopqrtuvwxyzABCDEFGHIJKLMNOPQRTUVWXYZ0123456789", k=len)
    return "".join(out)


class TestGeoseeqApiClient(TestCase):
    """Test suite for packet building."""

    def setUp(self):
        self.knex = Knex(ENDPOINT)
        # Creates a test user and an API token for the user in database. Returns the token.
        if TOKEN == "<no_token>":
            try:
                api_token = self.knex.post("/users/test-user",
                                           json={"email": f"clitestuser_{random_str()}@gmail.com"})
            except GeoseeqOtherError:
                print(f"Could not create test user on \"{ENDPOINT}\". If you are running this test suite "\
                      "against a live server, please set the GEOSEEQ_API_TOKEN environment variable to a "\
                      "valid API token.",
                      file=sys.stderr)
                raise
            except ConnectionError:
                print(f"Could not connect to GeoSeeq Server at \"{ENDPOINT}\".",
                      file=sys.stderr)
                raise
            self.knex.add_api_token(api_token)
        else:
            self.knex.add_api_token(TOKEN)
            try:
                me = self.knex.get("/users/me")  # Test that the token is valid
                self.username = me["name"]
                self.org_name = f"API_TEST_ORG 1 {self.username}"
            except GeoseeqNotFoundError:
                print(f"Could not connect to GeoSeeq Server at \"{ENDPOINT}\" with the provided token.  "\
                      "Is it possible that you set thd testing endpoint to a front end url instead of "\
                      "the corresponding backend url?",
                      file=sys.stderr)
                raise

    def test_create_org(self):
        """Test that we can create an org."""
        org = Organization(self.knex, self.org_name)
        org.idem()
        self.assertTrue(org.uuid)


    def test_create_project(self):
        """Test that we can create a project."""
        key = random_str()
        org = Organization(self.knex, self.org_name)

        proj = org.project(f"my_client_test_project {key}")
        proj.create()
        self.assertTrue(org.uuid)
        self.assertTrue(proj.uuid)

    def test_create_project_result_folder(self):
        """Test that we can create a result folder in a project."""
        key = random_str()
        org = Organization(self.knex, self.org_name)
        proj = org.project(f"my_client_test_proj {key}")
        # N.B. It should NOT be necessary to call <parent>.create()
        result_folder = proj.result_folder(f"my_client_test_module_name")  # no {key} necessary
        result_folder.create()
        self.assertTrue(org.uuid)
        self.assertTrue(proj.uuid)
        self.assertTrue(result_folder.uuid)

    def test_create_project_result_file(self):
        """Test that we can create a result file in a project."""
        key = random_str()
        org = Organization(self.knex, self.org_name)
        proj = org.project(f"my_client_test_proj {key}")
        result_folder = proj.result_folder(f"my_client_test_module_name")  # no {key} necessary
        # N.B. It should NOT be necessary to call <parent>.create()
        result_file = result_folder.result_file("my_client_test_field_name", {"foo": "bar"})
        result_file.create()
        self.assertTrue(org.uuid)
        self.assertTrue(proj.uuid)
        self.assertTrue(result_folder.uuid)
        self.assertTrue(result_file.uuid)

    def test_create_sample(self):
        """Test that we can create a sample."""
        key = random_str()
        org = Organization(self.knex, self.org_name)
        proj = org.project(f"my_client_test_proj {key}")
        # N.B. It should NOT be necessary to call <parent>.create()
        samp = proj.sample(f"my_client_test_sample {key}")
        samp.create()
        self.assertTrue(org.uuid)
        self.assertTrue(proj.uuid)
        self.assertTrue(samp.uuid)

    def test_add_sample(self):
        """Test that we can create a sample and add it to a different project."""
        key = random_str()
        org = Organization(self.knex, self.org_name)
        proj1 = org.project(f"my_client_test_proj1 {key}")
        samp = proj1.sample(f"my_client_test_sample {key}").create()

        proj2 = org.project(f"my_client_test_proj2 {key}").create()
        proj2.add_sample(samp).save()
        self.assertIn(samp.uuid, {samp.uuid for samp in proj2.get_samples()})

    def test_get_samples_project(self):
        """Test that we can get the samples in a project."""
        key = random_str()
        org = Organization(self.knex, self.org_name)
        proj = org.project(f"my_client_test_proj {key}")
        samp_names = [f"my_client_test_sample_{i} {key}" for i in range(10)]
        for samp_name in samp_names:
            proj.sample(samp_name).create()
        retrieved_proj = org.project(f"my_client_test_proj {key}").get()
        retrieved_names = set()
        for samp in retrieved_proj.get_samples():
            retrieved_names.add(samp.name)
            self.assertTrue(samp.uuid)
        for samp_name in samp_names:
            self.assertIn(samp_name, retrieved_names)

    def test_get_result_folders_in_project(self):
        """Test that we can get the result folders in a project."""
        key = random_str()
        org = Organization(self.knex, self.org_name)
        proj = org.project(f"my_client_test_proj {key}")
        result_names = [("my_client_test_module", f"replicate_{i}") for i in range(10)]
        for module_name, replicate in result_names:
            proj.result_folder(module_name, replicate=replicate).create()
        retrieved_proj = org.project(f"my_client_test_proj {key}").get()
        retrieved_names = set()
        for result in retrieved_proj.get_result_folders():
            retrieved_names.add((result.module_name, result.replicate))
            self.assertTrue(result.uuid)
        for result_name_rep in result_names:
            self.assertIn(result_name_rep, retrieved_names)

    def test_get_result_folders_in_sample(self):
        """Test that we can get the result folders in a sample."""
        key = random_str()
        org = Organization(self.knex, self.org_name)
        proj = org.project(f"my_client_test_proj {key}")
        samp = proj.sample(f"my_client_test_sample {key}").create()
        result_names = [("my_client_test_module", f"replicate_{i}") for i in range(10)]
        for module_name, replicate in result_names:
            samp.result_folder(module_name, replicate=replicate).create()
        retrieved = proj.sample(f"my_client_test_sample {key}").get()
        retrieved_names = set()
        for result in retrieved.get_result_folders():
            retrieved_names.add((result.module_name, result.replicate))
            self.assertTrue(result.uuid)
        for result_name_rep in result_names:
            self.assertIn(result_name_rep, retrieved_names)

    def test_get_result_files(self):
        """Test that we can get the files in a result folder."""
        key = random_str()
        org = Organization(self.knex, self.org_name)
        proj = org.project(f"my_client_test_proj {key}")
        samp = proj.sample(f"my_client_test_sample {key}")
        result_folder = samp.result_folder("my_client_test_module").create()
        self.assertTrue(proj.uuid)

        field_names = [f"field_{i}" for i in range(10)]
        for field_name in field_names:
            result_folder.field(field_name).create()

        retrieved = samp.result_folder("my_client_test_module").get()
        retrieved_names = set()
        for result in retrieved.get_fields():
            retrieved_names.add(result.name)
            self.assertTrue(result.uuid)
        for result_name_rep in field_names:
            self.assertIn(result_name_rep, retrieved_names)

    def test_modify_sample(self):
        """Test that we can modify a sample after creation"""
        key = random_str()
        org = Organization(self.knex, self.org_name)
        proj = org.project(f"my_client_test_proj {key}")
        # N.B. It should NOT be necessary to call <parent>.create()
        samp = proj.sample(f"my_client_test_sample {key}")
        samp.create()
        self.assertTrue(samp.uuid)
        self.assertTrue(samp._already_fetched)
        self.assertFalse(samp._modified)
        samp.metadata = {f"metadata_{key}": "some_new_metadata"}
        self.assertTrue(samp._modified)
        samp.save()
        self.assertTrue(samp._already_fetched)
        self.assertFalse(samp._modified)
        retrieved = proj.sample(f"my_client_test_sample {key}").get()
        self.assertIn(f"metadata_{key}", retrieved.metadata)

    def test_create_sample_result_folder(self):
        """Test that we can create a result folder in a sample."""
        key = random_str()
        org = Organization(self.knex, self.org_name)
        proj = org.project(f"my_client_test_proj {key}")
        samp = proj.sample(f"my_client_test_sample {key}")
        # N.B. It should NOT be necessary to call <parent>.create()
        result_folder = samp.result_folder(f"my_client_test_module_name")  # no {key} necessary
        result_folder.create()
        self.assertTrue(org.uuid)
        self.assertTrue(proj.uuid)
        self.assertTrue(samp.uuid)
        self.assertTrue(result_folder.uuid)

    def test_create_sample_result_file(self):
        """Test that we can create a result file in a sample."""
        key = random_str()
        org = Organization(self.knex, self.org_name)
        proj = org.project(f"my_client_test_proj {key}")
        samp = proj.sample(f"my_client_test_sample {key}")
        result_folder = samp.result_folder(f"my_client_test_module_name")  # no {key} necessary
        # N.B. It should NOT be necessary to call <parent>.create()
        result_file = result_folder.result_file("my_client_test_field_name", {"foo": "bar"})
        result_file.create()
        self.assertTrue(org.uuid)
        self.assertTrue(proj.uuid)
        self.assertTrue(samp.uuid)
        self.assertTrue(result_folder.uuid)
        self.assertTrue(result_file.uuid)

    @skip("failing on server (2023-10-24)")
    def test_modify_sample_result_file(self):
        """Test that we can modify a result file in a sample."""
        key = random_str()
        org = Organization(self.knex, self.org_name)
        proj = org.project(f"my_client_test_proj {key}")
        samp = proj.sample(f"my_client_test_sample {key}")
        result_folder = samp.result_folder(f"my_client_test_module_name")  # no {key} necessary
        # N.B. It should NOT be necessary to call <parent>.create()
        result_file = result_folder.result_file(f"my_client_test_file_name {key}", {"foo": "bar"})
        result_file.create()
        self.assertTrue(result_file.uuid)
        result_file.stored_data = {"foo": "bizz"}  # TODO: handle deep modifications
        result_file.save()
        retrieved = result_folder.result_file(f"my_client_test_file_name {key}").get()
        self.assertEqual(retrieved.stored_data["foo"], "bizz")

    def test_get_project_manifest(self):
        """Test that we can get a project manifest."""
        key = random_str()
        org = Organization(self.knex, self.org_name)
        proj = org.project(f"my_client_test_proj {key}")
        samp = proj.sample(f"my_client_test_sample {key}")
        result_folder = samp.result_folder(f"my_client_test_module_name")  # no {key} necessary
        result_file = result_folder.result_file("my_client_test_field_name", {"foo": "bar"})
        result_file.create()
        manifest = proj.get_manifest()
        self.assertTrue(manifest)

    def test_get_sample_manifest(self):
        """Test that we can get a sample manifest."""
        key = random_str()
        org = Organization(self.knex, self.org_name)
        proj = org.project(f"my_client_test_proj {key}")
        samp = proj.sample(f"my_client_test_sample {key}")
        result_folder = samp.result_folder(f"my_client_test_module_name")  # no {key} necessary
        result_file = result_folder.result_file("my_client_test_field_name", {"foo": "bar"})
        result_file.create()
        manifest = samp.get_manifest()
        self.assertTrue(manifest)



import click

from geoseeq import Organization, App
from geoseeq.id_constructors import resolve_id
from .shared_params import (
    use_common_state,
    project_id_arg,
    sample_ids_arg,
    yes_option,
    private_option,
    org_id_arg,
    handle_project_id,
    handle_multiple_sample_ids,
    handle_org_id,
)
from geoseeq.blob_constructors import org_from_uuid


@click.group('view')
def cli_view():
    """View objects on GeoSeeq."""
    pass


output_type_opt = click.option('--output-type', type=click.Choice(['uuid', 'name', 'all']), default='all', help='Type of output to print. Defaults to "all".')


def get_obj_output(obj, output_type):
    if output_type == 'uuid':
        return obj.uuid
    elif output_type == 'name':
        return obj.name
    else:
        return f'"{obj.name}"\t{obj.uuid}'


@cli_view.command('samples')
@use_common_state
@output_type_opt
@project_id_arg
def cli_list_samples(state, output_type, project_id):
    """Print a list of samples in the specified project.

    ---

    Example Usage:

    \b
    # List samples in "My Org/My Project"
    $ geoseeq view samples "My Org/My Project"

    ---

    Command Arguments:

    [PROJECT_ID] is the name or ID of the project to list samples from.

    ---
    """
    knex = state.get_knex()
    proj = handle_project_id(knex, project_id, create=False)
    for sample in proj.get_samples():
        print(get_obj_output(sample, output_type), file=state.outfile)


@cli_view.command('organizations')
@use_common_state
@output_type_opt
def cli_list_organizations(state, output_type):
    """Print a list of organizations.

    ---

    Example Usage:

    \b
    # List all organizations
    $ geoseeq view organizations

    ---
    """
    knex = state.get_knex()
    for uuid in Organization.all_uuids(knex):
        if output_type == 'uuid':
            print(uuid, file=state.outfile)
        else:
            org = org_from_uuid(knex, uuid)
            print(get_obj_output(org, output_type), file=state.outfile)


@cli_view.command('projects')
@use_common_state
@output_type_opt
@org_id_arg
def cli_list_projects(state, output_type, org_id):
    """Print a list of projects in the specified organization.

    ---

    Example Usage:

    \b
    # List projects in "My Org"
    $ geoseeq view projects "My Org"

    ---

    Command Arguments:

    [ORG_ID] is the name or ID of the organization to list projects from.

    ---
    """
    knex = state.get_knex()
    org = handle_org_id(knex, org_id, create=False)
    for proj in org.get_projects():
        print(get_obj_output(proj, output_type), file=state.outfile)


@cli_view.command('app')
@use_common_state
@click.argument('uuid')
def cli_view_app(state, uuid):
    """Print the specified app.

    ---

    Example Usage:

    \b
    # Print the app with UUID "d051ce05-f799-4aa7-8d8f-5cbf99136543"
    $ geoseeq view app d051ce05-f799-4aa7-8d8f-5cbf99136543

    ---

    Command Arguments:

    [UUID] is the UUID of the app to print.

    ---
    """
    knex = state.get_knex()
    app = App(knex, uuid)
    app.get()
    print(app)
    for field_name, field_value, optional in app.get_remote_fields():
        optional = "Optional" if optional else "Required"
        print(f'\t{field_name} :: "{field_value}" ({optional})')


@cli_view.command('project')
@use_common_state
@project_id_arg
def cli_view_project(state, project_id):
    """Print the specified project.

    ---

    Example Usage:

    \b
    # Print the project with ID "My Org/My Project"
    $ geoseeq view project "My Org/My Project"

    ---

    Command Arguments:

    [PROJECT_ID] is the name or ID of the project to print.

    ---
    """
    knex = state.get_knex()
    proj = handle_project_id(knex, project_id, create=False)
    print(proj)
    for field_name, field_value, optional in proj.get_remote_fields():
        optional = "Optional" if optional else "Required"
        print(f'\t{field_name} :: "{field_value}" ({optional})')


@cli_view.command('sample')
@use_common_state
@project_id_arg
@sample_ids_arg
def cli_view_sample(state, project_id, sample_ids):
    """Print the specified sample.

    ---

    Example Usage:

    \b
    # Print the sample with ID "My Sample" from "My Org/My Project/
    $ geoseeq view sample "My Org/My Project" "My Sample"

    ---

    Command Arguments:

    [PROJECT_ID] is the name or ID of the project to print the sample from.

    [SAMPLE_IDS]... is the name or ID of the sample to print.

    ---
    """
    knex = state.get_knex()
    proj = handle_project_id(knex, project_id, create=False)
    sample_ids = handle_multiple_sample_ids(knex, sample_ids, proj)
    for sample in sample_ids:
        print(sample)
        for field_name, field_value, optional in sample.get_remote_fields():
            optional = "Optional" if optional else "Required"
            print(f'\t{field_name} :: "{field_value}" ({optional})')


@cli_view.command('object')
@use_common_state
@click.argument('ids', nargs=-1)
def cli_view_object(state, ids):
    """Print the specified object as well as its type.

    ---

    Example Usage:

    \b
    # Print the object with name "My Org/My Project/My Sample/My Result Folder"
    $ geoseeq view object "My Org/My Project/My Sample/My Result Folder"

    \b
    # Print the object with GRN "grn:geoseeq:sample::d051ce05-f799-4aa7-8d8f-5cbf99136543"
    $ geoseeq view object "grn:geoseeq:sample::d051ce05-f799-4aa7-8d8f-5cbf99136543"

    ---

    Command Arguments:

    [ID] is the name or GRN of the object to print. UUIDs cannot be resolved without a type.

    ---
    """
    knex = state.get_knex()
    for id in ids:
        obj_type, obj = resolve_id(knex, id)
        print(f'{obj_type}:\t{obj}', file=state.outfile)


import json
import logging
import os
import time
import urllib.request
from os.path import basename, getsize, join
from pathlib import Path
from tempfile import NamedTemporaryFile

import requests

from geoseeq.constants import FIVE_MB
from geoseeq.remote_object import RemoteObject, RemoteObjectError
from geoseeq.utils import download_ftp, md5_checksum

from .bioinfo import SampleBioInfoFolder
from .result_file import ProjectResultFile, SampleResultFile
from .utils import *


class ResultFolder(RemoteObject):
    remote_fields = [
        "uuid",
        "created_at",
        "updated_at",
        "module_name",
        "replicate",
        "description",
        "is_private",
    ]

    @property
    def name(self):
        return self.module_name

    def _get(self, allow_overwrite=False):
        """Fetch the result from the server."""
        self.parent.idem()
        logger.debug(f"Getting AnalysisResult.")
        blob = self.get_cached_blob()
        if not blob:
            url = self.nested_url()
            if self.replicate:
                url += f"?replicate={self.replicate}"
            blob = self.knex.get(url, url_options=self.inherited_url_options)
            self.load_blob(blob, allow_overwrite=allow_overwrite)
            self.cache_blob(blob)
        else:
            self.load_blob(blob)

    def pre_hash(self):
        key = self.module_name + self.parent.pre_hash()
        key += self.replicate if self.replicate else ""
        return key

    def copy(self, new_parent, save=True):
        copied = new_parent.analysis_result(
            self.module_name, replicate=self.replicate, metadata=self.metadata
        )
        for field in self.get_fields():
            field.copy(copied, save=save)
        if save:
            copied.idem()
        return copied

AnalysisResult = ResultFolder # for backwards compatibility


class SampleResultFolder(ResultFolder, SampleBioInfoFolder):
    parent_field = "sample"

    def __init__(self, knex, sample, module_name, replicate=None, metadata={}, is_private=False):
        super().__init__(self)
        self.knex = knex
        self.sample = sample
        self.parent = self.sample
        self.module_name = module_name
        self.replicate = replicate
        self._get_field_cache = []
        self.metadata = metadata
        self.is_private = is_private

    def nested_url(self):
        return self.sample.nested_url() + f"/analysis_results/{self.module_name}"

    def _save(self):
        data = {field: getattr(self, field) for field in self.remote_fields if hasattr(self, field)}
        data["sample"] = self.sample.uuid
        url = f"sample_ars/{self.uuid}"
        d = {"data": data, "url": url, "sample_ar": self}
        logger.debug(f"Saving SampleAnalysisResult. {d}")
        self.knex.put(url, json=data, url_options=self.inherited_url_options)

    def get_post_data(self):
        """Return a dict that can be used to POST this result to the server."""
        data = {
            field: getattr(self, field)
            for field in self.remote_fields
            if hasattr(self, field) and getattr(self, field) is not None
        }
        data["sample"] = self.sample.uuid
        if self.replicate:
            data["replicate"] = self.replicate
        return data

    def _create(self):
        self.sample.idem()
        data = self.get_post_data()
        d = {"data": data, "sample_ar": self}
        logger.debug(f"Creating SampleAnalysisResult. {d}")
        blob = self.knex.post(
            f"sample_ars?format=json", json=data, url_options=self.inherited_url_options
        )
        self.load_blob(blob)

    def result_file(self, field_name, pipeline_run=None, data={}):
        d = {
            "data": data,
            "field_name": field_name,
            "pipeline_run": pipeline_run,
            "sample_ar": self,
        }
        logger.debug(f"Creating SampleAnalysisResultField for SampleAnalysisResult. {d}")
        return SampleResultFile(self.knex, self, field_name, pipeline_run=pipeline_run, data=data)

    def field(self, *args, **kwargs):
        return self.result_file(*args, **kwargs)

    def get_result_files(self, cache=True):
        """Return a list of ar-fields fetched from the server."""
        if cache and self._get_field_cache:
            for field in self._get_field_cache:
                yield field
            return
        url = f"sample_ar_fields?analysis_result_id={self.uuid}"
        # url = self.nested_url() + f"/fields"
        logger.debug(f"Fetching SampleAnalysisResultFields. {self}")
        result = self.knex.get(url)
        for result_blob in result["results"]:
            result = self.field(result_blob["name"])
            result.load_blob(result_blob)
            # We just fetched from the server so we change the RemoteObject
            # meta properties to reflect that
            result._already_fetched = True
            result._modified = False
            if cache:
                self._get_field_cache.append(result)
            else:
                yield result
        if cache:
            for field in self._get_field_cache:
                yield field

    def get_fields(self, *args, **kwargs):
        return self.get_result_files(*args, **kwargs)

    def __str__(self):
        return f"<Geoseeq::SampleResultFolder {self.module_name} {self.replicate} {self.uuid} />"

SampleAnalysisResult = SampleResultFolder # for backwards compatibility


class ProjectResultFolder(ResultFolder):
    parent_field = "grp"

    def __init__(self, knex, grp, module_name, replicate=None, metadata={}, is_private=False):
        super().__init__(self)
        self.knex = knex
        self.grp = grp
        self.parent = self.grp
        self.module_name = module_name
        self.replicate = replicate
        self.metadata = metadata
        self.is_private = is_private

    def nested_url(self):
        return self.grp.nested_url() + f"/analysis_results/{self.module_name}"

    def _save(self):
        data = {field: getattr(self, field) for field in self.remote_fields if hasattr(self, field)}
        data["sample_group"] = self.grp.uuid
        url = f"sample_group_ars/{self.uuid}"
        self.knex.put(url, json=data)

    def _create(self):
        self.grp.idem()
        data = {
            "sample_group": self.grp.uuid,
            "module_name": self.module_name,
        }
        if self.replicate:
            data["replicate"] = self.replicate
        if self.uuid:
            data["uuid"] = self.uuid
        blob = self.knex.post(f"sample_group_ars?format=json", json=data)
        self.load_blob(blob)

    def result_file(self, field_name, pipeline_run=None, data={}):
        return ProjectResultFile(self.knex, self, field_name, pipeline_run=pipeline_run, data=data)

    def field(self, *args, **kwargs):
        return self.result_file(*args, **kwargs)

    def get_result_files(self):
        """Return a list of ar-fields fetched from the server."""
        url = f"sample_group_ar_fields?analysis_result_id={self.uuid}"
        result = self.knex.get(url)
        for result_blob in result["results"]:
            result = self.field(result_blob["name"])
            result.load_blob(result_blob)
            # We just fetched from the server so we change the RemoteObject
            # meta properties to reflect that
            result._already_fetched = True
            result._modified = False
            yield result

    def get_fields(self, *args, **kwargs):
        return self.get_result_files(*args, **kwargs)

    def __str__(self):
        return f"<Geoseeq::ProjectResultFolder {self.module_name} {self.replicate} {self.uuid} />"

SampleGroupAnalysisResult = ProjectResultFolder # for backwards compatibility



from .result import SampleResultFolder
from .remote_object import RemoteObject


class Sample(RemoteObject):
    remote_fields = [
        "uuid",
        "created_at",
        "updated_at",
        "name",
        "metadata",
        "library",
        "description",
    ]
    parent_field = "lib"
    url_prefix = "samples"

    def __init__(self, knex, lib, name, metadata={}):
        super().__init__(self)
        self.knex = knex
        self.lib = lib
        self.new_lib = None
        self.name = name
        self.metadata = metadata
        self._get_result_cache = []

    @property
    def project(self):
        return self.lib

    @property
    def brn(self):
        return f'brn:{self.knex.instance_code()}:sample:{self.uuid}'

    def nested_url(self):
        return self.lib.nested_url() + f"/samples/{self.name}"

    def change_library(self, new_lib):
        self.new_lib = new_lib
        self._modified = True

    def _save(self):
        data = self.get_post_data()
        url = f"samples/{self.uuid}"
        self.knex.put(url, json=data, url_options=self.inherited_url_options)
        if self.new_lib:
            self.lib = self.new_lib
            self.new_lib = None

    def _get(self, allow_overwrite=False):
        """Fetch the result from the server."""
        self.lib.get()
        blob = self.get_cached_blob()
        if not blob:
            url = self.nested_url()
            blob = self.knex.get(url, url_options=self.inherited_url_options)
            self.load_blob(blob, allow_overwrite=allow_overwrite)
            self.cache_blob(blob)
        else:
            self.load_blob(blob, allow_overwrite=allow_overwrite)

    def get_post_data(self):
        data = {field: getattr(self, field) for field in self.remote_fields if hasattr(self, field)}
        data["library"] = self.lib.uuid
        if self.new_lib:
            if isinstance(self.new_lib, RemoteObject):
                data["library"] = self.new_lib.uuid
            else:
                data["library"] = self.new_lib
        if data['uuid'] is None:
            data.pop('uuid')
        return data

    def _create(self):
        assert self.lib.is_library
        self.lib.idem()
        data = self.get_post_data()
        url = f"samples"
        blob = self.knex.post(url, json=data, url_options=self.inherited_url_options)
        self.load_blob(blob)

    def delete(self):
        url = f"samples/{self.uuid}"
        self.knex.delete(url)
        self._already_fetched = False
        self._deleted = True

    def result_folder(self, module_name, replicate=None, metadata=None):
        """Return a SampleResultFolder for this sample."""
        return SampleResultFolder(
            self.knex, self, module_name, replicate=replicate, metadata=metadata
        )

    def analysis_result(self, *args, **kwargs):
        """Return a SampleResultFolder for this sample.
        
        This is an alias for result_folder."""
        return self.result_folder(*args, **kwargs)
    
    def get_result_folders(self, cache=True):
        """Yield sample analysis results fetched from the server."""
        self.get()
        if cache and self._get_result_cache:
            for ar in self._get_result_cache:
                yield ar
            return
        url =  f"sample_ars?sample_id={self.uuid}"
        result = self.knex.get(url)
        for result_blob in result["results"]:
            result = self.analysis_result(result_blob["module_name"])
            result.load_blob(result_blob)
            # We just fetched from the server so we change the RemoteObject
            # meta properties to reflect that
            result._already_fetched = True
            result._modified = False
            if cache:
                self._get_result_cache.append(result)
            else:
                yield result
        if cache:
            for ar in self._get_result_cache:
                yield ar

    def get_analysis_results(self, cache=True):
        """Yield sample analysis results fetched from the server.
        
        This is an alias for get_result_folders.
        """
        return self.get_result_folders(cache=cache)

    def get_manifest(self):
        """Return a manifest for this sample."""
        url = f"samples/{self.uuid}/manifest"
        return self.knex.get(url)

    def __str__(self):
        return f"<Geoseeq::Sample {self.name} {self.uuid} />"

    def __repr__(self):
        return f"<Geoseeq::Sample {self.name} {self.uuid} />"

    def pre_hash(self):
        return "SAMPLE" + self.name + self.lib.pre_hash()

    def copy(self, sample_group, save=True):
        copied = sample_group.sample(self.name, self.metadata)
        for ar in self.get_analysis_results():
            ar.copy(copied, save=save)
        if save:
            copied.idem()
        return copied



import uuid


def is_grn(el):
    """Return True if `el` is a GeoSeeq Resource Number (GRN)"""
    return el.startswith('grn:')


def is_uuid(el):
    """Return True if `el` is a UUID"""
    try:
        uuid.UUID(el)
        return True
    except ValueError:
        return False


def is_name(el):
    """Return True if `el` is a name.
    
    e.g. "My Org/My Project"
    """
    if "/" in el:
        # every name except org names have a slash
        return True
    if not is_grn_or_uuid(el):
        # if the name has no slash and is not a grn or uuid, it's a name
        return True
    return False


def is_grn_or_uuid(el):
    """Return True if `el` is a GRN or a UUID"""
    return is_grn(el) or is_uuid(el)
